//
//  CatalogSong.swift
//  CatalogSong
//
//  Created by Rudrank Riyam on 14/08/21.
//

import MusicKit

/// A collection of songs.
public typealias Songs = MusicItemCollection<Song>

public extension MusadoraKit {
  /// Fetch a song from the Apple Music catalog by using its identifier.
  /// - Parameters:
  ///   - id: The unique identifier for the song.
  ///   - properties: Additional relationships to fetch with the song.
  /// - Returns: `Song` matching the given identifier.
  static func catalogSong(id: MusicItemID,
                          with properties: [PartialMusicAsyncProperty<Song>] = []) async throws -> Song {
    var request = MusicCatalogResourceRequest<Song>(matching: \.id, equalTo: id)
    request.properties = properties
    let response = try await request.response()

    guard let song = response.items.first else {
      throw MusadoraKitError.notFound(for: id.rawValue)
    }
    return song
  }

  /// Fetch a song from the Apple Music catalog by using its identifier with all properties.
  /// - Parameters:
  ///   - id: The unique identifier for the song.
  /// - Returns: `Song` matching the given identifier.
  static func catalogSong(id: MusicItemID) async throws -> Song {
    var request = MusicCatalogResourceRequest<Song>(matching: \.id, equalTo: id)
    request.properties = .all
    let response = try await request.response()

    guard let song = response.items.first else {
      throw MusadoraKitError.notFound(for: id.rawValue)
    }
    return song
  }

  /// Fetch multiple songs from the Apple Music catalog by using their identifiers.
  /// - Parameters:
  ///   - ids: The unique identifiers for the songs.
  ///   - properties: Additional relationships to fetch with the songs.
  /// - Returns: `Songs` matching the given identifiers.
  static func catalogSongs(ids: [MusicItemID],
                           with properties: [PartialMusicAsyncProperty<Song>] = []) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.id, memberOf: ids)
    request.properties = properties
    let response = try await request.response()
    return response.items
  }

  /// Fetch multiple songs from the Apple Music catalog by using their identifiers with all properties.
  /// - Parameters:
  ///   - ids: The unique identifiers for the songs.
  /// - Returns: `Songs` matching the given identifiers.
  static func catalogSongs(ids: [MusicItemID]) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.id, memberOf: ids)
    request.properties = .all
    let response = try await request.response()
    return response.items
  }

  /// Fetch one or more songs from Apple Music catalog by using their ISRC value.
  ///
  /// Note that one ISRC value may return more than one song.
  /// - Parameters:
  ///   - isrc: The ISRC values for the songs.
  ///   - properties: Additional relationships to fetch with the songs.
  /// - Returns: `Songs` matching the given ISRC value.
  static func catalogSong(isrc: String,
                          with properties: [PartialMusicAsyncProperty<Song>] = []) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.isrc, equalTo: isrc)
    request.properties = properties
    let response = try await request.response()
    return response.items
  }

  /// Fetch one or more songs from Apple Music catalog by using their ISRC value with all properties.
  ///
  /// Note that one ISRC value may return more than one song.
  /// - Parameters:
  ///   - isrc: The ISRC values for the songs.
  /// - Returns: `Songs` matching the given ISRC value.
  static func catalogSong(isrc: String) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.isrc, equalTo: isrc)
    request.properties = .all
    let response = try await request.response()
    return response.items
  }

  /// Fetch multiple songs from Apple Music catalog by using their ISRC values.
  ///
  /// Note that one ISRC value may return more than one song.
  /// - Parameters:
  ///   - isrc: The ISRC values for the songs.
  ///   - properties: Additional relationships to fetch with the songs.
  /// - Returns: `Songs` matching the given ISRC values.
  static func catalogSongs(isrc: [String],
                           with properties: [PartialMusicAsyncProperty<Song>] = []) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.isrc, memberOf: isrc)
    request.properties = properties
    let response = try await request.response()
    return response.items
  }

  /// Fetch multiple songs from Apple Music catalog by using their ISRC values with all properties.
  ///
  /// Note that one ISRC value may return more than one song.
  /// - Parameters:
  ///   - isrc: The ISRC values for the songs.
  /// - Returns: `Songs` matching the given ISRC values.
  static func catalogSongs(isrc: [String]) async throws -> Songs {
    var request = MusicCatalogResourceRequest<Song>(matching: \.isrc, memberOf: isrc)
    request.properties = .all
    let response = try await request.response()
    return response.items
  }
}

extension Array where Element == PartialMusicAsyncProperty<Song> {
  public static var all: Self {
    var properties: [PartialMusicAsyncProperty<Song>] = [.albums, .artists, .composers, .genres, .musicVideos, .artistURL, .station]
#if compiler(>=5.7)
    if #available(iOS 16.0, macOS 13.0, tvOS 16.0, watchOS 9.0, *) {
      properties += [.audioVariants]
      return properties
    } else {
      return properties
    }
#else
    return properties
#endif
  }
}
